import logging
import time
from tornado import gen


logger = logging.getLogger(__name__)


class DefaultSleepPolicy(object):
    BASE_SLEEP_INTERVAL = 0.5
    MAX_SLEEP_INTERVAL = 3 * 60.
    SLEEP_REPEAT_COUNT = 10

    def __init__(self, sleep_interval=BASE_SLEEP_INTERVAL,
                 max_sleep_interval=MAX_SLEEP_INTERVAL,
                 sleep_repeat_count=SLEEP_REPEAT_COUNT, **kwargs):

        self._base_sleep_interval = self._sleep_interval = sleep_interval
        self._max_sleep_interval = max_sleep_interval
        self._sleep_repeat_count = self._base_sleep_repeat_count = sleep_repeat_count

    def _next_sleep_interval(self):
        if not self._sleep_repeat_count:
            self._sleep_repeat_count = self._base_sleep_repeat_count
            self._sleep_interval *= 2
            if self._sleep_interval > self._max_sleep_interval:
                self._sleep_interval = self._max_sleep_interval
            return
        self._sleep_repeat_count -= 1

    def sleep(self):
        f = gen.sleep(self._sleep_interval)
        f.add_done_callback(lambda _: self._next_sleep_interval())
        return f

    @property
    def cur_interval(self):
        return self._sleep_interval

    def reset(self):
        self._sleep_interval = self._base_sleep_interval
        self._sleep_repeat_count = self._base_sleep_repeat_count


class GreedyPolicy(object):
    MAX_COUNT = 100
    TIMEOUT = 10.0

    def __init__(self, max_count=MAX_COUNT, timeout=TIMEOUT, **kwargs):

        self._max_count = max_count
        self._count = 0
        self._timeout = timeout
        self._time = time.time()

    def need_to_wait(self):
        self._count += 1
        return (self._count >= self._max_count
                or self._time + self._timeout < time.time())

    def wait(self):
        self._count = 0
        self._time = time.time()
        return gen.moment



import logging
from tornado import gen
from tornado.concurrent import Future

from ..config import Config
from ..msg import Msg
from ..utils import mark_as_proxy_method, is_proxy, simple_oid


logger = logging.getLogger(__name__)


class BaseRole(object):
    defaults = Config(
        unique=False,  # means, the object will be singleton / shared
        start_delay=0,
    )
    _singleton_cache = {}

    @classmethod
    def factory(cls, *args, **kwargs):
        unique = kwargs.get('unique', False) or cls.defaults.unique
        if unique:
            # if unique, than behave like a singleton
            key = cls.__singleton_key__(args, kwargs)
            if key in cls._singleton_cache:
                return cls._singleton_cache[key]
            new_obj = cls(*args, **kwargs)
            cls._singleton_cache[key] = new_obj
            return new_obj

        return cls(*args, **kwargs)

    @classmethod
    def __singleton_key__(cls, passed_args, passed_kwargs):
        return u'{}.{}:{}'.format(cls.__module__,
                                  cls.__name__, passed_kwargs.get('config_name', 'unknown'))

    @property
    def is_unique(self):
        return self.config.unique

    @property
    def is_active(self):
        return self.started \
               or self.config.namespace_default in self.namespaces \
               or self.config.namespace_default in self.app.active_namespaces \
               or bool(set(self.namespaces).intersection(self.app.active_namespaces))

    def __init__(self, *items, **config):
        self._oid = simple_oid()

        self.app = config.pop('app')
        self.parent = config.pop('parent')
        self.namespaces = config.pop('namespaces', None) or self.app.namespaces
        self.config = self.defaults.copy_and_update(config)
        self.items = items

        self._started = self._unique_start_lock = False
        self._started_futures = []
        self._stopped_futures = []
        self.input = self.output = None
        self.send = getattr(self, 'send', None)
        self._forward = getattr(self, '_forward', None)

    def __str__(self):
        return u'{}:{}'.format(self.__class__.__name__, self.parent)

    @property
    def active_items(self):
        return [i for i in self.items if i.is_active]

    def construct_subrole(self, name, conf):
        if isinstance(conf, (list, tuple)):
            conf = {'*': conf}
        elif conf is None:
            conf = {}
        conf['app'] = self.app
        conf['parent'] = self
        return self.app.config.find_and_construct_class(name=name, kwargs=conf)

    @property
    def started(self):
        return self._started

    @started.setter
    def started(self, value):
        self._started = bool(value)
        futures = self._started_futures if self.started else self._stopped_futures
        for future in futures:
            future.set_result(self._started)
        if self._started:
            logger.debug('[%s] Started.', self)

    def wait_for_start(self):
        f = Future()
        if not self.started:
            self._started_futures.append(f)
        else:
            f.set_result(True)
        return f

    def wait_for_stop(self):
        f = Future()
        if not self.started:
            self._stopped_futures.append(f)
        else:
            f.set_result(False)
        return f

    def link_methods(self):
        if getattr(self, 'send', None) is None:
            self.send = self.get_send_method()

        if getattr(self, '_forward', None) is None:
            self._forward = self.get_forward_method()

    @mark_as_proxy_method
    def _input_forwarder(self, data):
        return self.output.send(data)

    def get_send_method(self):
        return getattr(self, 'on_recv', self._input_forwarder)

    def get_forward_method(self):
        method = self._input_forwarder
        obj = self.output

        while is_proxy(method):
            if obj:
                method = obj.send
                obj = obj.output
            else:
                break

        return method

    def set_input(self, obj):
        self.input = obj

    def set_output(self, obj):
        self.output = obj

    def _prepare_message_meta(self, **extra):
        extra.setdefault('host', None)
        return extra

    def _prepare_message(self, data):
        return Msg(
            message=data,
            source=None,
            meta=self._prepare_message_meta()
        )

    def _pre_start(self):
        pass

    def _post_stop(self):
        pass

    @gen.coroutine
    def start(self):
        need_to_skip_start = self.started or not self.is_active
        if not need_to_skip_start and self.is_unique and self._unique_start_lock:
            need_to_skip_start = True

        if not need_to_skip_start:
            if self.is_unique:
                self._unique_start_lock = True

            if self.config.start_delay:
                yield gen.sleep(self.config.start_delay)

            logger.debug('[%s] Starting...', self)
            if hasattr(self.output, 'wait_for_start'):
                yield gen.maybe_future(self.output.wait_for_start())
            yield gen.maybe_future(self._pre_start())

            # mark obj as 'started'
            self.started = True

            # notify that the obj will be reused for other pipes in case of uniqueness
            if self.is_unique:
                logger.debug('[%s] Started in a shared mode.', self)

    @gen.coroutine
    def stop(self):
        if self.started:
            logger.debug('[%s] Stopping...', self)
            if hasattr(self.input, 'wait_for_stop'):
                yield self.input.wait_for_stop()
            self.started = False
            yield gen.maybe_future(self._post_stop())



import anydbm
import pickle
import os


class Register(object):

    def __init__(self, index_file):
        self._index_file = os.path.expandvars(os.path.expanduser(index_file))
        if not os.path.exists(os.path.dirname(self._index_file)):
            os.makedirs(os.path.dirname(self._index_file))
        self._reg = anydbm.open(self._index_file, 'c')

    def set(self, key, val):
        self._reg[key] = val

    def get(self, key):
        return self._reg[key]

    def __getitem__(self, item):
        return self._reg[item]

    def get_path(self, name):
        return pickle.loads(self.get(name))

    def set_path(self, path_obj):
        self.set(path_obj.name, pickle.dumps(path_obj))


#!/usr/bin/env python
from distutils.core import setup
import os

from logdog import (
    __version__ as version,
    __author__ as author,
    __email__ as author_email)


requires = [
    r.strip() for r in
    open(os.path.join(os.getcwd(), 'requirements.txt')).readlines()
]

setup(
    name='logdog',
    version=version,
    description='logs watching + webui',
    author=author,
    author_email=author_email,
    url='https://github.com/miphreal/python-logdog/',
    packages=['logdog'],
    scripts=['scripts/logdog'],
    install_requires=requires
)


import glob
from itertools import chain
import logging
import os
from tornado import gen

from logdog.core.config import Config, handle_as_list
from logdog.core.path import Path
from logdog.core.register import Register


logger = logging.getLogger(__name__)


class Application(object):

    def __init__(self, active_namespaces, config, io_loop=None,
                 force_handler=None, force_sources=None):
        from tornado.ioloop import IOLoop
        self.io_loop = io_loop or IOLoop.current()
        self.config = config
        self.force_handler = force_handler
        self.force_sources = force_sources
        self.namespaces = (Config.namespace_default,)
        self.active_namespaces = active_namespaces or [Config.namespace_default]
        logger.debug('[%s] Active namespaces: %s', self, ', '.join(self.active_namespaces))
        self._pipes = {}
        self._sources = {}
        self.register = Register(index_file=config.options.sources.index_file)

    def __str__(self):
        return u'APP'

    @gen.coroutine
    def load_sources(self):
        flat_files = set()
        flat_patterns = set()

        sources = self.config.sources
        if self.force_sources:
            self.force_sources = self.force_sources.split(':')
            sources = handle_as_list(self.force_sources)

        for source, conf in sources:
            if not isinstance(source, (list, tuple)):
                source = [source]
            else:
                source = list(source)

            if isinstance(conf, basestring):
                conf = Config(handler=conf)
            elif not conf:
                conf = Config(handler=self.config.options.sources.default_handler)
            elif isinstance(conf, dict):
                conf.setdefault('handler', self.config.options.sources.default_handler)
            else:
                logger.warning('[APP] Weird config for "%s" (will be skipped).', ', '.join(source))
                continue

            if self.force_handler:
                conf['handler'] = self.force_handler

            intersection_patterns = flat_patterns.intersection(source)
            if intersection_patterns:
                logger.warning('[APP] Duplicate source patterns: %s (will be skipped).',
                               ', '.join(intersection_patterns))
                source = list(set(source).difference(intersection_patterns))

            source.sort()
            source = tuple(source)

            files = chain(*map(glob.glob, source))
            files = set(filter(os.path.isfile, files))

            intersection_files = flat_files.intersection(files)
            if intersection_files:
                logger.warning('[APP] Your source patterns have intersections.'
                               'The following files appeared in several groups: %s',
                               ', '.join(intersection_files))

                # remove from wider pattern
                for source_, (_, files_) in self.register.iteritems():
                    intersection_ = files_.intersection(files)
                    if intersection_ and len(files_) > len(files):
                        files_.difference_update(intersection_)
                        logger.warning('[APP] "%s" will not be a part of "%s" group.',
                                       ', '.join(intersection_), ', '.join(source_))
                    elif intersection_:
                        files.difference_update(intersection_)
                        logger.warning('[APP] "%s" will not be a part of "%s" group.',
                                       ', '.join(intersection_), ', '.join(source))

            flat_files.update(files)
            flat_patterns.update(source)
            self._sources[source] = (conf, files)

    @gen.coroutine
    def construct_pipes(self):
        default_watcher = self.config.options.sources.default_watcher
        for conf, files in self._sources.itervalues():
            conf['app'] = self
            conf['parent'] = self

            for f in files:
                watcher = self.config.find_and_construct_class(
                    name=conf.get('watcher', default_watcher), kwargs=conf)
                pipe = self.config.find_and_construct_class(name=conf['handler'], kwargs=conf)

                try:
                    path = self.register.get_path(f)
                except KeyError:
                    path = Path(f, 0, None)

                watcher.set_input(path)
                watcher.set_output(pipe)
                pipe.link_methods()
                watcher.link_methods()

                self._pipes[f] = watcher, pipe

    @gen.coroutine
    def _init(self):
        yield self.load_sources()
        yield self.construct_pipes()

        pipes = [p for _, p in self._pipes.itervalues()]
        watchers = [w for w, _ in self._pipes.itervalues()]
        yield [i.start() for i in chain(pipes, watchers)]

    def run(self):
        self.io_loop.add_callback(self._init)
        self.io_loop.start()



# == POLLER ==  # periodically getting and forwarding data
# def poll():  -- active polling
#   while True:
#       self.on_recv(input.pop())
# def on_recv(data):  -- passive waiting for messages
#   output.on_recv(data)

# == COLLECTOR ==  # waiting for data and collecting
# def on_recv(data):
#   (storage|buffer).put(data)
# def pop():
#   return (storage|buffer).pop()

# == PROCESSOR ==  # accepting data, processing and forwarding
# def on_recv(data):
#   .. process data
#   output.on_recv(data)

# == CONNECTOR(SENDER, RECEIVER) ==  # transfer messages over network
# def SENDER.on_recv(data):
#   <socket>.send(data)
# def RECEIVER.on_recv(data):
#   output.on_recv(data)

# == VIEWER ==  # accepting data and displaying it
# def on_recv(data):
#   .. show some representation of data
